# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# https://stackoverflow.com/questions/63475461/unable-to-import-opengl-gl-in-python-on-macos

""" Hack for macOS Big Sur """
from ctypes import util


def patch_ctypes():
    orig_util_find_library = util.find_library

    def new_util_find_library(name):
        res = orig_util_find_library(name)
        if res:
            return res
        return f'/System/Library/Frameworks/{name}.framework/{name}'

    util.find_library = new_util_find_library



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import struct
import sys
import threading
from numpy import dot, cross, array, zeros, cos, sin, uint8, uint32
from numpy.linalg import norm
import spynnaker_visualisers.opengl_support as gl
import spynnaker_visualisers.glut_framework as glut


class RaytraceDrawer(glut.GlutFramework):
    __slots__ = (
        "_moving", "_strafing", "_turn_down", "_turn_right", "_rolling",
        "_height", "_width", "_win_height", "_win_width",
        "_viewing_frame", "_received_frame", "_sockfd_input",
        "_look", "_up", "_position")
    moveAmount = 0.00003
    turnAmount = 0.0000003

    # Fields of view
    VERT_FOV = 50.0
    HORIZ_FOV = 60.0

    INPUT_PORT_SPINNAKER = 17894
    SDP_HEADER = struct.Struct("<HBBBBHHHHIII")
    PIXEL_FORMAT = struct.Struct(">HHBBB")
    RECV_BUFFER_SIZE = 1500  # Ethernet MTU; SpiNNaker doesn't jumbo

    def __init__(self, size=256):
        super().__init__()
        self._moving = 0
        self._strafing = 0
        # Turn left is negative
        self._turn_right = 0
        # Turn up is negative
        self._turn_down = 0
        self._rolling = 0
        self._position = array([-220.0, 50.0, 0.0])
        self._look = array([1.0, 0.0, 0.0])
        self._up = array([0.0, 1.0, 0.0])
        self._height = size
        self._width = int(self.HORIZ_FOV * self._height / self.VERT_FOV)
        self._win_height = self._height
        self._win_width = self._width
        self._viewing_frame = zeros(
            self._width * self._height * 3, dtype=uint8)
        self._received_frame = zeros(
            self._width * self._height, dtype=uint32)
        self._sockfd_input = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sockfd_input.bind(('0.0.0.0', self.INPUT_PORT_SPINNAKER))

    def start(self, args):
        threading.Thread(target=self._input_thread, daemon=True).start()
        self.start_framework(
            args, "Path Tracer", self._width, self._height, 0, 0, 10,
            display_mode=glut.displayModeDouble)

    def init(self):
        gl.enable(gl.blend, gl.depth_test)
        gl.blend_function(gl.src_alpha, gl.one_minus_src_alpha)

    def display(self, dTime):
        gl.clear_color(1.0, 1.0, 1.0, 0.001)
        # pylint: disable=unsupported-binary-operation
        gl.clear(gl.color_buffer_bit | gl.depth_buffer_bit)
        gl.draw_pixels(
            self._win_width, self._win_height, gl.rgb, gl.unsigned_byte,
            self._viewing_frame.data)

    def reshape(self, width, height):
        self._win_width = min(width, self._width)
        self._win_height = min(height, self._height)
        gl.viewport(0, 0, width, height)
        gl.load_identity()

    def special_keyboard_down(self, key, x, y):
        if key == glut.keyUp:
            self._turn_down = -1
        elif key == glut.keyDown:
            self._turn_down = 1
        elif key == glut.keyRight:
            self._rolling = -1
        elif key == glut.keyLeft:
            self._rolling = 1

    def special_keyboard_up(self, key, x, y):
        if key == glut.keyUp or key == glut.keyDown:
            self._turn_down = 0
        elif key == glut.keyLeft or key == glut.keyRight:
            self._rolling = 0

    def keyboard_down(self, key, x, y):
        if key == 'w':
            self._moving = 1
        elif key == 's':
            self._moving = -1
        elif key == 'a':
            self._turn_right = -1
        elif key == 'd':
            self._turn_right = 1
        elif key == 'q':
            self._strafing = 1
        elif key == 'e':
            self._strafing = -1
        elif key == '\x1b':  # Escape
            sys.exit()

    def keyboard_up(self, key, x, y):
        if key == 'w' or key == 's':
            self._moving = 0
        elif key == 'a' or key == 'd':
            self._turn_right = 0
        elif key == 'q' or key == 'e':
            self._strafing = 0

    @staticmethod
    def vector_rotate(rotated, axis, theta):
        """Rotate the first vector around the second"""
        # https://gist.github.com/fasiha/6c331b158d4c40509bd180c5e64f7924
        par = (dot(rotated, axis) / dot(axis, axis) * axis)
        perp = rotated - par
        w = cross(axis, perp)
        w = w / norm(w)
        result = par + perp * cos(theta) + norm(perp) * w * sin(theta)
        return result / norm(result)

    def calculate_movement(self, dt):
        # Forward movement
        if self._moving:
            self._position += self._look * dt * self.moveAmount * self._moving
        right = cross(self._up, self._look)
        # Strafing movement
        if self._strafing:
            self._position += right * dt * self.moveAmount * self._strafing
        # To turn left/right, rotate the look vector around the up vector
        if self._turn_right:
            self._look = self.vector_rotate(
                self._look, self._up, dt * self.turnAmount * self._turn_right)
        # To turn up/down, rotate the look vector and up vector about the right
        # vector
        if self._turn_down:
            self._look = self.vector_rotate(
                self._look, right, dt * self.turnAmount * self._turn_down)
            self._up = self.vector_rotate(
                self._up, right, dt * self.turnAmount * self._turn_down)
        # To roll, rotate the up vector around the look vector
        if self._rolling:
            self._up = self.vector_rotate(
                self._up, self._look, dt * self.turnAmount * self._rolling)

    def run(self):
        """Calculate movement ten times a second"""
        super().run()
        self.calculate_movement(self.frame_time_elapsed * 1000)

    def _input_thread(self):
        print(
            f"Drawer running (listening port: {self.INPUT_PORT_SPINNAKER})...")
        while True:
            msg = self._sockfd_input.recv(self.RECV_BUFFER_SIZE)
            sdp_msg = self.SDP_HEADER.unpack_from(msg)
            data = msg[self.SDP_HEADER.size:]  # sdp_msg.data
            if sdp_msg[7] == 3:  # sdp_msg.command
                for pixel_datum in self._pixelinfo(
                        data, sdp_msg[9]):  # sdp_msg.arg1
                    self.process_one_pixel(*pixel_datum)

    @classmethod
    def _pixelinfo(cls, data, number_of_pixels):
        for i in range(number_of_pixels):
            yield cls.PIXEL_FORMAT.unpack_from(
                data, i * cls.PIXEL_FORMAT.size)

    def process_one_pixel(self, x, y, r, g, b):
        index = (self._height - y - 1) * self._width + x
        if index < self._width * self._height:
            ix3 = index * 3
            count = self._received_frame[index]
            cp1 = count + 1
            self._viewing_frame[ix3] = (
                (r + count * self._viewing_frame[ix3]) // cp1)
            self._viewing_frame[ix3 + 1] = (
                (g + count * self._viewing_frame[ix3 + 1]) // cp1)
            self._viewing_frame[ix3 + 2] = (
                (b + count * self._viewing_frame[ix3 + 2]) // cp1)
            self._received_frame[index] += 1


def main(args):
    drawer = RaytraceDrawer()
    drawer.start(args)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# encoding: utf-8
""" A live plotter for the sPyNNaker Sudoku network.
"""
from argparse import ArgumentParser, REMAINDER
import sys
from threading import Condition, RLock
from spinn_utilities.overrides import overrides
from spinn_front_end_common.utilities.connections import LiveEventConnection
from spynnaker_visualisers.glut_framework import GlutFramework
from spynnaker_visualisers.opengl_support import (
    vertex, draw, lines, color, point_size, points, line_width, clear_color,
    clear, color_buffer_bit, load_identity, viewport, matrix_mode, projection,
    model_view, orthographic_projction, shade_model, smooth)

__all__ = []
__version__ = 1
__date__ = '2017-07-25'

WINDOW_BORDER = 110
INIT_WINDOW_WIDTH = 800
INIT_WINDOW_HEIGHT = 600
INIT_WINDOW_X = 100
INIT_WINDOW_Y = 100
FRAMES_PER_SECOND = 10


class SudokuPlot(GlutFramework):
    """ A live plotter for the sPyNNaker Sudoku network.
    """
    __slots__ = [
        "args",
        "cell_id",
        "cell_labels",
        "cell_size_map",
        "database_read",
        "label_to_cell_map",
        "latest_time",
        "ms_per_bin",
        "n_neurons",
        "n_populations_to_read",
        "neurons_per_number",
        "plot_time_ms",
        "point_mutex",
        "points_to_draw",
        "simulation_started",
        "start_condition",
        "timestep_ms",
        "user_pressed_start",
        "window_height",
        "window_width"]

    def __init__(self, args, neurons_per_number, ms_per_bin, wait_for_start):
        """
        :param args:
            Arguments (relating to the display) to pass through to GLUT
        :param neurons_per_number:
            How many neurons are used per number in the Sudoku cells
        :param ms_per_bin:
            How long does a sampling period last
        :param wait_for_start:
            Whether the system should wait for the SpiNNaker simulation to\
            boot (probably yes!)
        """
        super(SudokuPlot, self).__init__()
        self.window_width = INIT_WINDOW_WIDTH
        self.window_height = INIT_WINDOW_HEIGHT

        self.cell_id = 0
        self.user_pressed_start = not wait_for_start
        self.simulation_started = False
        self.database_read = False
        self.n_neurons = 0
        self.timestep_ms = 0
        self.plot_time_ms = 0
        self.ms_per_bin = float(ms_per_bin)
        self.latest_time = 0.0
        self.neurons_per_number = neurons_per_number

        self.n_populations_to_read = 1

        self.args = args

        self.points_to_draw = [[] for _ in range(81)]
        self.point_mutex = RLock()

        self.label_to_cell_map = dict()
        self.cell_size_map = dict()
        self.cell_labels = dict()

        self.start_condition = Condition()

    @overrides(GlutFramework.init)
    def init(self):
        clear_color(0.0, 0.0, 0.0, 1.0)
        color(1.0, 1.0, 1.0)
        shade_model(smooth)

    def connect_callbacks(self, connection, label):
        """ Arrange so that labels on the given connection report their\
            goings-on to this class.

        :type connection: LiveEventConnection
        :type label: str
        """
        connection.add_init_callback(label, self._init_cb)
        connection.add_receive_callback(label, self._receive_cb)
        connection.add_start_resume_callback(label, self._start_cb)

    def _init_cb(self, label, n_neurons, run_time_ms, machine_time_step_ms):
        self.plot_time_ms = float(run_time_ms)
        self.timestep_ms = float(machine_time_step_ms)

        self.cell_labels[self.cell_id] = label
        self.cell_size_map[self.cell_id] = n_neurons
        self.label_to_cell_map[label] = self.cell_id
        self.n_neurons += n_neurons
        self.cell_id += 1

        with self.start_condition:
            self.n_populations_to_read -= 1
            if self.n_populations_to_read <= 0:
                self.database_read = True
                while not self.user_pressed_start:
                    self.start_condition.wait()

    def _start_cb(self, *args):
        with self.start_condition:
            self.simulation_started = True

    def _receive_cb(self, label, time, spikes=None):
        if spikes is None:
            spikes = []
        with self.point_mutex:
            for spike in spikes:
                cell_id, neuron_id = divmod(
                    spike, self.neurons_per_number * 9)
                self.points_to_draw[cell_id].append((time, neuron_id))
            time_ms = time * self.timestep_ms
            if time_ms > self.latest_time:
                self.latest_time = time_ms

    def main_loop(self):
        """ Run the GUI.
        """
        self.start_framework(
            self.args, "Sudoku", self.window_width, self.window_height,
            INIT_WINDOW_X, INIT_WINDOW_Y, FRAMES_PER_SECOND)

    @overrides(GlutFramework.display)
    def display(self, dTime):
        self._start_display()

        cell_width = (self.window_width - 2 * WINDOW_BORDER) / 9.0
        cell_height = (self.window_height - 2 * WINDOW_BORDER) / 9.0
        end = self.latest_time
        start = end - self.ms_per_bin
        if start < 0.0:
            start = 0.0
            end = start + self.ms_per_bin

        with self.start_condition:
            if not self.database_read:
                prompt = "Waiting for simulation to load..."
            elif not self.user_pressed_start:
                prompt = "Press space bar to start..."
            elif not self.simulation_started:
                prompt = "Waiting for simulation to start..."
            else:
                prompt = "Sudoku"
        self._print_text(prompt)

        self._draw_cells(cell_width, cell_height)

        if self.timestep_ms != 0:
            x_spacing = cell_width / ((end - start) / self.timestep_ms)
            start_tick = int(start / self.timestep_ms)
            with self.point_mutex:
                values, probs = self._find_cell_values(start_tick)
                valid = self._find_cell_correctness(values)
                self._draw_cell_contents(values, valid, probs, start_tick,
                                         x_spacing, cell_width, cell_height)

    @overrides(GlutFramework.reshape)
    def reshape(self, width, height):
        self.window_width = width
        self.window_height = height

        # Viewport dimensions
        viewport(0, 0, width, height)
        matrix_mode(projection)
        load_identity()

        # An orthographic projection. Should probably look into OpenGL
        # perspective projections for 3D if that's your thing
        orthographic_projction(0.0, width, 0.0, height, -50.0, 50.0)
        matrix_mode(model_view)
        load_identity()

    @overrides(GlutFramework.keyboard_down)
    def keyboard_down(self, key, x, y):
        if key == 32 or key == ' ':
            with self.start_condition:
                if not self.user_pressed_start:
                    print("Starting the simulation")
                    self.user_pressed_start = True
                    self.start_condition.notify_all()

    def _find_cell_values(self, start_tick):
        cell_value = [0] * 81
        cell_prob = [0.0] * 81

        for cell in range(81):
            # Strip off items that are no longer needed
            queue = self.points_to_draw[cell]
            while queue and queue[0][0] < start_tick:
                queue.pop(0)

            # Count the spikes per number
            count, total = self._count_spikes_per_number(queue)

            # Work out the probability of a given number in a given cell
            max_prob_number = 0
            max_prob = 0.0
            for i in range(9):
                if count[i] > 0:
                    prob = count[i] / total
                    if prob > max_prob:
                        max_prob = prob
                        max_prob_number = i + 1
            cell_value[cell] = max_prob_number
            cell_prob[cell] = max_prob
        return cell_value, cell_prob

    def _count_spikes_per_number(self, queue):
        count = [0] * 9
        total = 0
        for (_, n_id) in queue:
            number = n_id // self.neurons_per_number
            if number < 9:
                count[number] += 1
                total += 1
            else:
                sys.stderr.write(f"Neuron id {n_id} out of range\n")
        return count, float(total)

    def _find_cell_correctness(self, values):
        # Work out the correctness of each cell
        cell_valid = [True] * 81
        for cell in range(81):
            y, x = divmod(cell, 9)
            for row in range(9):
                if row != y:
                    self._check_cell(values, cell_valid, x, y, row, x)
            for col in range(9):
                if col != x:
                    self._check_cell(values, cell_valid, x, y, y, col)
            for row in range(3 * (y // 3), 3 * (y // 3 + 1)):
                for col in range(3 * (x // 3), 3 * (x // 3 + 1)):
                    if x != col and y != row:
                        self._check_cell(values, cell_valid, x, y, row, col)
        return cell_valid

    @staticmethod
    def _start_display():
        point_size(1.0)
        clear(color_buffer_bit)
        clear_color(1.0, 1.0, 1.0, 1.0)
        color(0.0, 0.0, 0.0, 1.0)

    # TODO positioning
    # https://github.com/SpiNNakerManchester/sPyNNakerVisualisers/issues/23
    def _print_text(self, prompt):
        # Guesstimate of length of prompt in pixels
        plen = len(prompt) * 4
        self.write_large(
            self.window_width / 2 - plen, self.window_height - 50, prompt)

    def _draw_cells(self, width, height):
        color(0.0, 0.0, 0.0, 1.0)
        for i in range(10):
            line_width(3.0 if i % 3 == 0 else 1.0)
            pos = WINDOW_BORDER + i * height
            self._line(self.window_width - WINDOW_BORDER, pos,
                       WINDOW_BORDER, pos)
            pos = WINDOW_BORDER + i * width
            self._line(pos, self.window_height - WINDOW_BORDER,
                       pos, WINDOW_BORDER)

    def _draw_cell_contents(self, value, valid, prob, start, x_spacing,
                            cell_width, cell_height):
        # Print the spikes
        for cell in range(81):
            cell_y, cell_x = divmod(cell, 9)
            x_start = WINDOW_BORDER + (cell_x * cell_width) + 1
            y_start = WINDOW_BORDER + (cell_y * cell_height) + 1
            y_spacing = cell_height / (self.neurons_per_number * 9.0)

            # Work out how probable the number is and use this for colouring
            cell_sat = 1 - prob[cell]

            point_size(2.0)
            with draw(points):
                if valid[cell]:
                    color(cell_sat, 1.0, cell_sat, 1.0)
                else:
                    color(1.0, cell_sat, cell_sat, 1.0)
                for (time, n_id) in self.points_to_draw[cell]:
                    x_value = (time - start) * x_spacing + x_start
                    y_value = n_id * y_spacing + y_start
                    vertex(x_value, y_value)

            # Print the number
            if value[cell] != 0:
                color(0, 0, 0, 1 - cell_sat)
                size = 0.005 * cell_height
                self.write_small(
                    x_start + (cell_width / 2.0) - (size * 50.0),
                    y_start + (cell_height / 2.0) - (size * 50.0),
                    size, 0, "%d", value[cell])

    @staticmethod
    def _line(x1, y1, x2, y2):
        with draw(lines):
            vertex(x1, y1)
            vertex(x2, y2)

    @staticmethod
    def _check_cell(values, correct, x, y, row, col):
        value = values[y * 9 + x]
        if value == values[row * 9 + col]:
            correct[y * 9 + x] = False


# https://github.com/SpiNNakerManchester/sPyNNakerVisualisers/issues/24
def sudoku_visualiser(args, port=19999, neurons=5, ms=100, database=None):
    """ Make a visualiser, connecting a LiveEventConnection that listens to a\
        population labelled "Cells" to a GLUT GUI.
    """
    # Set up the application
    cells = ["Cells"]
    connection = LiveEventConnection(
        "LiveSpikeReceiver", receive_labels=cells, local_port=port)
    plotter = SudokuPlot(args, neurons, ms, database is None)
    for label in cells:
        plotter.connect_callbacks(connection, label)
    if database is not None:
        # TODO: This concept not present on Python side!
        # connection.set_database(database)
        sys.stderr.write("Database setting not currently supported")
    plotter.main_loop()


def main(argv=None):
    """ The main script.\
        Parses command line arguments and launches the visualiser.
    """
    program_name = "sudoku_visualiser"
    program_version = "v%d" % (__version__)
    program_description = "Visualise the SpiNNaker sudoku solver."
    program_version_string = '%%prog %s (%s)' % (program_version, __date__)

    # setup option parser
    parser = ArgumentParser(prog=program_name,
                            description=program_description)
    parser.add_argument(
        "-d", "--database", dest="database", metavar="FILE",
        help="optional file path to where the database is located, if "
        "needed for manual configuration", default=None)
    parser.add_argument(
        "-m", "--ms_per_bin", dest="ms", metavar="MILLISECONDS",
        help="optional number of milliseconds to show at once",
        type=float, default=100)
    parser.add_argument(
        "-n", "--neurons_per_number", dest="neurons",
        help="the number of neurons that represent each number in a cell",
        metavar="COUNT", type=int, default=5)
    parser.add_argument(
        "-p", "--hand_shake_port", dest="port", default="19999",
        help="optional port which the visualiser will listen to for"
        " database hand shaking", metavar="PORT", type=int)
    parser.add_argument('--version', action='version',
                        version=program_version_string)
    parser.add_argument("args", nargs=REMAINDER)

    # Set up and run the application
    try:
        if argv is None:
            argv = sys.argv[1:]
        sudoku_visualiser(**parser.parse_args(argv).__dict__)
        return 0
    except Exception as e:  # pylint: disable=broad-except
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2


if __name__ == "__main__":
    sys.exit(main())



"""
A basic framework that implements the basic system handling for a GUI
application that uses OpenGL and GLUT to do the GUI work.
"""
#
# Copyright (c) 2017 The University of Manchester
# Based on work Copyright (c) 2010 Paul Solt, PaulSolt@gmail.com
# Originally released under the MIT License
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @author: Converted to Python by Donal Fellows

from datetime import datetime
import os
import traceback
import OpenGL.error
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker_visualisers.opengl_support import (
    viewport, save_matrix, enable, blend, line_smooth, disable, line_width,
    blend_function, src_alpha, one_minus_src_alpha, rotate, scale, translate,
    raster_position)
try:
    # this fails in <=2020 versions of Python on OS X 11.x
    import OpenGL.GLUT  # noqa: F401
except ImportError:
    # Hack for macOS Big Sur
    from ._bigsurhack import patch_ctypes
    patch_ctypes()
import OpenGL.GLUT as GLUT

keyUp = GLUT.GLUT_KEY_UP
keyDown = GLUT.GLUT_KEY_DOWN
keyLeft = GLUT.GLUT_KEY_LEFT
keyRight = GLUT.GLUT_KEY_RIGHT
displayModeDouble = GLUT.GLUT_DOUBLE


class _PerformanceTimer(object):
    __slots__ = [
        "_stamp_1", "_stamp_2", "_stopped"]

    @staticmethod
    def _now():
        return datetime.now()

    def __init__(self):
        self._stopped = True
        self._stamp_1 = 0
        self._stamp_2 = 0

    def start(self):
        """ Start the timer. """
        self._stopped = False
        self._stamp_1 = _PerformanceTimer._now()

    def stop(self):
        """ Stop the timer. """
        self._stamp_2 = _PerformanceTimer._now()
        self._stopped = True

    @property
    def stopped(self):
        """ Is the timer stopped? """
        return self._stopped

    @property
    def elapsed_milliseconds(self):
        """ How long elapsed in the last timing run? In milliseconds.

        ..note::
            Only valid when the timer has previously been run and is currently\
            stopped.
        """
        delta = self._stamp_2 - self._stamp_1
        return float(delta.seconds) * 1000 + float(delta.microseconds) / 1000

    @property
    def elapsed_seconds(self):
        """ How long elapsed in the last timing run? In seconds.

        ..note::
            Only valid when the timer has previously been run and is currently\
            stopped.
        """
        delta = self._stamp_2 - self._stamp_1
        return float(delta.seconds) + float(delta.microseconds) / 1000000


class GlutFramework(object, metaclass=AbstractBase):
    ''' Base for code that wants to visualise using an OpenGL surface.
    '''
    # pylint: disable=broad-except
    __slots__ = [
        "display_timer",
        "elapsed_time_in_seconds",
        "frame_rate_timer",
        "frame_time",
        "frame_time_elapsed",
        "_logged_errors",
        "window"]

    def __init__(self):
        self.window = None
        self.frame_time_elapsed = 0.0
        self.frame_time = 0.0
        self.frame_rate_timer = _PerformanceTimer()
        self.display_timer = _PerformanceTimer()
        self.elapsed_time_in_seconds = 0.0
        self._logged_errors = set()

    # pylint: disable=unsupported-binary-operation
    def start_framework(self, args, title, width, height, posx, posy, fps, *,
                        display_mode=GLUT.GLUT_RGB | GLUT.GLUT_DOUBLE):
        """ start_framework will initialize framework and start the GLUT run\
            loop. It must be called after the GlutFramework class is created\
            to start the application.

        Not expected to return.
        """
        # Sets the instance to this, used in the callback wrapper functions
        self.frame_time = 1.0 / fps * 1000.0

        # Initialize GLUT
        GLUT.glutInit(args)
        GLUT.glutInitDisplayMode(display_mode)
        GLUT.glutInitWindowSize(width, height)
        GLUT.glutInitWindowPosition(posx, posy)
        self.window = GLUT.glutCreateWindow(title)
        try:
            GLUT.glutSetOption(GLUT.GLUT_ACTION_ON_WINDOW_CLOSE,
                               GLUT.GLUT_ACTION_CONTINUE_EXECUTION)
        except OpenGL.error.NullFunctionError:
            pass

        self.init()  # Initialize

        # Function callbacks with wrapper functions
        GLUT.glutDisplayFunc(self.__display_framework)
        GLUT.glutReshapeFunc(self.__reshape_framework)
        GLUT.glutIdleFunc(self.__run)
        GLUT.glutMouseFunc(self.__mouse_button_press)
        GLUT.glutMotionFunc(self.__mouse_move)
        GLUT.glutKeyboardFunc(self.__keyboard_down)
        GLUT.glutKeyboardUpFunc(self.__keyboard_up)
        GLUT.glutSpecialFunc(self.__special_keyboard_down)
        GLUT.glutSpecialUpFunc(self.__special_keyboard_up)
        try:
            GLUT.glutCloseFunc(self._terminate)
        except OpenGL.error.NullFunctionError:
            GLUT.glutWMCloseFunc(self._terminate)

        GLUT.glutMainLoop()

    def init(self):
        """ Initialises GLUT and registers any extra callback functions.
        """

    @abstractmethod
    def display(self, dTime):
        """ The display function is called at a specified frames-per-second\
            (FPS). Any animation drawing code can be run in the display method.

        :param dTime: the change in time (seconds)
        """

    def reshape(self, width, height):
        """ Called when the window dimensions change.

        :param width: the width of the window in pixels
        :param height: the height of the window in pixels
        """
        viewport(0, 0, width, height)

    def mouse_button_press(self, button, state, x, y):
        """ Called when the mouse buttons are pressed.

        :param button: the mouse buttons
        :param state: the state of the buttons
        :param x: the x coordinate
        :param y: the y coordinate
        """

    def mouse_move(self, x, y):
        """ Called when the mouse moves on the screen.

        :param x: the x coordinate
        :param y: the y coordinate
        """

    def keyboard_down(self, key, x, y):
        """ The keyboard function is called when a standard key is pressed\
            down.

        :param key: the key press
        :param x: the x coordinate of the mouse
        :param y: the y coordinate of the mouse
        """

    def keyboard_up(self, key, x, y):
        """ The keyboard function is called when a standard key is "unpressed".

        :param key: the key press
        :param x: the x coordinate of the mouse
        :param y: the y coordinate of the mouse
        """

    def special_keyboard_down(self, key, x, y):
        """ The keyboard function is called when a special key is pressed down\
            (F1 keys, Home, Inser, Delete, Page Up/Down, End, arrow keys).\
            https://www.opengl.org/resources/libraries/glut/spec3/node54.html

        :param key: the key press
        :param x: the x coordinate of the mouse
        :param y: the y coordinate of the mouse
        """

    def special_keyboard_up(self, key, x, y):
        """ The keyboard function is called when a special key is "unpressed"\
            (F1 keys, Home, Inser, Delete, Page Up/Down, End, arrow keys).

        :param key: the key press
        :param x: the x coordinate of the mouse
        :param y: the y coordinate of the mouse
        """

    def run(self):
        """ The run method is called by GLUT and contains the logic to set the\
            frame rate of the application.
        """
        if self.frame_rate_timer.stopped:
            self.frame_rate_timer.start()

        # stop the timer and calculate time since last frame
        self.frame_rate_timer.stop()
        milliseconds = self.frame_rate_timer.elapsed_milliseconds
        self.frame_time_elapsed += milliseconds

        if self.frame_time_elapsed >= self.frame_time:
            # If the time exceeds a certain "frame rate" then show the next
            # frame
            GLUT.glutPostRedisplay()

            # remove a "frame" and start counting up again
            self.frame_time_elapsed -= self.frame_time
        self.frame_rate_timer.start()

    def display_framework(self):
        """ The display_framework() function sets up initial GLUT state and\
            calculates the change in time between each frame. It calls the\
            display(float) function, which can be subclassed.
        """
        if self.display_timer.stopped:
            self.display_timer.start()
        self.display_timer.stop()
        elapsedTimeInSeconds = self.display_timer.elapsed_seconds
        if GLUT.glutGetWindow() == self.window:
            self.display(elapsedTimeInSeconds)
            GLUT.glutSwapBuffers()
        self.display_timer.start()

    def reshape_framework(self, width, height):
        """ Handle resizing of the window.
        """
        if GLUT.glutGetWindow() == self.window:
            self.reshape(width, height)

    @staticmethod
    def write_large(x, y, string, *args):
        """ Utility function: write a string to a given location as a bitmap.
        """
        # pylint: disable=no-member
        if args:
            string = string % args
        raster_position(x, y)
        for ch in string:
            GLUT.glutBitmapCharacter(GLUT.GLUT_BITMAP_TIMES_ROMAN_24, ord(ch))

    @staticmethod
    def write_small(x, y, size, rotation, string, *args):
        """ Utility function: write a string to a given location as a strokes.
        """
        # pylint: disable=no-member
        if args:
            string = string % args

        with save_matrix():
            # antialias the font
            enable(blend, line_smooth)
            blend_function(src_alpha, one_minus_src_alpha)
            line_width(1.5)

            translate(x, y, 0.0)
            scale(size, size, size)
            rotate(rotation, 0.0, 0.0, 1.0)
            for ch in string:
                GLUT.glutStrokeCharacter(GLUT.GLUT_STROKE_ROMAN, ord(ch))
            disable(blend, line_smooth)

    @staticmethod
    def _terminate(exit_code=0):
        """
        Because sys.exit() doesn't always work in the ctype-handled callbacks.
        """
        os._exit(exit_code)  # pylint: disable=protected-access

    def __display_framework(self):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.display_framework()
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __reshape_framework(self, width, height):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.reshape_framework(width, height)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __run(self):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.run()
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __mouse_button_press(self, button, state, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.mouse_button_press(button, state, x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __mouse_move(self, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.mouse_move(x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __keyboard_down(self, key, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.keyboard_down(key.decode(), x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __keyboard_up(self, key, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.keyboard_up(key.decode(), x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __special_keyboard_down(self, key, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.special_keyboard_down(key, x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __special_keyboard_up(self, key, x, y):
        if not GLUT.glutGetWindow():
            return
        try:
            return self.special_keyboard_up(key, x, y)
        except Exception:
            self.__log_error()
        except SystemExit:
            self._terminate()

    def __log_error(self):
        tb = traceback.format_exc()
        if tb not in self._logged_errors:
            self._logged_errors.add(tb)
            traceback.print_exc()



# Copyright (c) 2018 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file makes the OpenGL interface a little more python-pretty. It's
massively incomplete; feel free to add to it as required.
"""

from contextlib import contextmanager
try:
    # this fails in <=2020 versions of Python on OS X 11.x
    import OpenGL.GL  # noqa: F401  # pylint: disable=unused-import
except ImportError:
    # Hack for macOS Big Sur
    from ._bigsurhack import patch_ctypes
    patch_ctypes()
import OpenGL.GL as GL

# pylint: disable=invalid-name
blend = GL.GL_BLEND
color_buffer_bit = GL.GL_COLOR_BUFFER_BIT
depth_buffer_bit = GL.GL_DEPTH_BUFFER_BIT
line_smooth = GL.GL_LINE_SMOOTH
lines = GL.GL_LINES
model_view = GL.GL_MODELVIEW
one_minus_src_alpha = GL.GL_ONE_MINUS_SRC_ALPHA
points = GL.GL_POINTS
projection = GL.GL_PROJECTION
smooth = GL.GL_SMOOTH
src_alpha = GL.GL_SRC_ALPHA
depth_test = GL.GL_DEPTH_TEST
rgb = GL.GL_RGB
unsigned_byte = GL.GL_UNSIGNED_BYTE
# pylint: enable=invalid-name


def blend_function(sfactor, dfactor):
    """ Set the blending function. """
    GL.glBlendFunc(sfactor, dfactor)


def clear(mask):
    """ Clear the drawing surface. """
    GL.glClear(mask)


def clear_color(red, green, blue, alpha=1.0):
    """ Clear the surface to the given colour. """
    GL.glClearColor(float(red), float(green), float(blue), float(alpha))


def color(*args):
    """ Set the drawing colour. """
    GL.glColor(*args)


def disable(*args):
    """ Disable the listed features. """
    for feature in args:
        GL.glDisable(feature)


def enable(*args):
    """ Enable the listed features. """
    for feature in args:
        GL.glEnable(feature)


def line_width(width):
    """ Set the line width. """
    GL.glLineWidth(float(width))


def load_identity():
    """ Load the identity matrix. """
    GL.glLoadIdentity()


def matrix_mode(mode):
    """ Set the matrix mode. """
    GL.glMatrixMode(mode)


def orthographic_projction(*args):
    """ Set an orthographic (non-perspective) projection. """
    GL.glOrtho(*args)


def point_size(size):
    """ Set the size of points. """
    GL.glPointSize(float(size))


def raster_position(*args):
    """ Set the raster position. """
    GL.glRasterPos(*args)


def rotate(angle, x, y, z):
    """ Rotate the projection about a point. """
    GL.glRotatef(angle, x, y, z)


def scale(x, y, z):
    """ Scale the projection about the origin. """
    GL.glScale(x, y, z)


def shade_model(mode):
    """ Set the shading model. """
    GL.glShadeModel(mode)


def translate(x, y, z):
    """ Translate the projection. """
    GL.glTranslate(x, y, z)


def vertex(*args):
    """ Mark a vertex of a drawing path. """
    GL.glVertex(*args)


def viewport(x, y, width, height):
    """ Set up the view port. """
    GL.glViewport(int(x), int(y), int(width), int(height))


def draw_pixels(*args):
    GL.glDrawPixels(*args)


@contextmanager
def draw(drawing_style):
    """ Draw a line, set of points or closed curve (depending on\
        drawing_style). Use as a context manager and specify the vertices of\
        the path in the body of the context.
    """
    GL.glBegin(drawing_style)
    yield
    GL.glEnd()


@contextmanager
def save_matrix():
    """ Manipulate the view matrix in a temporary context; the view matrix is\
        restored once this context is left.
    """
    GL.glPushMatrix()
    yield
    GL.glPopMatrix()
